###############################################################################
#
# Ferret search engine
#
# Written by Brian C. White <bcwhite@pobox.com>
# Copyright (c) 1996-2003 by Brian White and Behan Webster
#
###############################################################################

=head1 NAME

ferret - "Ferret" search engine Perl interface

=cut

package Ferret;

require 5.002;
require Exporter;
require DynaLoader;
require AutoLoader;

use Carp;

@ISA = qw(Exporter DynaLoader AutoLoader);
$VERSION   = '2.5';
@EXPORT	   = qw();
@EXPORT_OK = qw();

#$] == ===REQUIRED-PERL=== or die "Error: This instance of Ferret has been compiled against Perl v===REQUIRED-PERL===\n		  The currently installed version of Perl is v$]\n";

bootstrap Ferret $VERSION;



sub AUTOLOAD {
	my $constname = $AUTOLOAD;
	$constname =~ s/.*:://;
	my $val = constant($constname, @_ ? $_[0] : 0);
	if ($! != 0) {
#		if ($! =~ /Invalid/) {
			$AutoLoader::AUTOLOAD = $AUTOLOAD;
			goto &AutoLoader::AUTOLOAD;
#		} else {
#		  Carp::croak("Error: Nonexistent Ferret constant '$constname' used");
#		}
	}
	eval "sub $AUTOLOAD { $val }";
	goto &$AUTOLOAD;
}



use strict;



=head1 SYNOPSIS

    use Ferret;

    [result =] Ferret::routine([...parameters...]);

      or

    $ferret = new Ferret;
    [result =] $ferret->method(...parameters...);


Routines:
===ROUTINES===

Methods:
===METHODS===

=head1 DESCRIPTION

The B<ferret> library provides an easy way to build advanced indexing and
query functions using Perl.

=head1 QUERY STRUCTURE SUMMARY

This section provides some help on doing searches using Ferret.  If you are
new to Ferret, you may want to skip the summary (the table below), and read
the section called "Query Structure Details" first, and then use the Summary
section as a reference table.

The following table summarizes all the operations that Ferret can perform.
The "Quick Example" column shows how the operation is to be entered.  More
detailed explanation of the operation follow after the table.

Please read the few notes that follow the table; they are important.


    +=================================================+
    |                        |                        |
    | Operation              | Quick Example(s)       |
    |                        |                        |
    +=================================================+
    |                        |                        |
    | Single Word            | word                   |
    |                        |                        |
    | Negation               | NOT word               |
    |                        | ! word                 |
    |                        |                        |
    | Required Words (AND)   | word AND werd AND wird |
    |                        | word & werd & wird     |
    |                        | word werd wird         |
    |                        |                        |
    | Wild Characters        | w?rd                   |
    |                        | w*                     |
    |                        | w[oei]rd               |
    |                        |                        |
    | Optional Words (OR)    | word OR werd OR wird   |
    |                        | word | werd | wird     |
    |                        |                        |
    | Priority (Parentheses) | word & (werd | wird)   |
    |                        |                        |
    | Proximity              | word NEAR werd         |
    |                        | word 10 werd           |
    |                        |                        |
    | Phrase                 | "werd is the word"     |
    |                        | 1 werd is the word     |
    |                        |                        |
    +------------------------+------------------------+

=over 4

=item Non-Alphabetical Characters:

Ferret operates only on alphabetical characters.  You cannot perform searches
for numbers.  All characters other than alphabetical (numbers, symbols,
punctuation) are ignored by the actual search engine; they are treated the
same as spaces.  However, when searching for phrases, including them can be
useful because the search results will be displayed properly.

=item Case Sensetivity:

If a word is typed in all lowercase letters, the search on it is
case-insensetive.  If there is at least one capital letter, the search is
case-sensetive.  The exception to this rule are the special keywords,
described next.

=item Special Keywords:

The words B<AND>, B<OR>, B<NOT>, and B<NEAR> (all capital letters) are special
keywords.  B<AND> has the same meaning as the B<&> operator; B<OR> has the
same meaining as the B<|> operator; B<NOT> has the same meaning as the B<!>
operator; and B<NEAR> has the same meaning as a proximity-number of B<10>.

=item Order of Operations:

All operations take place left to right; certain operators do not take
priority over others.  Therefore,

    first OR second AND third

will select documents containing "third," and one of the words "first" or
"second."  It does not select documents containing either "first" or both the
words "second" and "third."

=back

=head1 QUERY STRUCTURE DETAILS

The following gives a detailed description of B<Ferret>'s query language and
longer examples of how it can be used.

=head2 Single Word Search

To retrieve documents containing a single, specific word, just type the word
by itself:

    word

The word must not contain punctuation, numbers, or other special characters.
Please see the note on non-alphabetic characters in the previous section.

As stated earlier, if the word is all lowercase, then the search is not
case-sensetive.  If there is at least one capital in the word, then the search
is case sensetive.

=head2 Multi-Word Search

Being able to join queries of multiple words is the primary goal of any search
engine.

=over 4

=item Required Word Set (AND Operation)

To retrieve documents containing each and every word of a set, type:

    first AND second AND third

(The B<AND> keyword must be all uppercase.)  Or, type the words separated by
ampersands:

    first & second & third

or, the above may be abbreviated to:

    first second third

=item Optional Word Set (OR Operation)

To retrieve documents containing at least one of a set of words:

    first OR second OR third

(The B<OR> keyword must be all uppercase.)  Or,

    first | second | third

=back

As stated earlier, the B<AND> ( B<&> ) operation does not take priority over
the B<OR> ( B<|> ) operation; priority is strictly left to right, unless you
use parentheses.

=head2 Prioritization

All evaluation by ferret occurs left to right, regardless of the operation.
To prioritize sections of a query, parentheses are required.

For example, to retrieve documents containing some required words and at least
one of several optional words:

    firstreqd & secordreqd & (firstoptl | secondoptl)

=head2 Negation

The negation operator reverses the meaning of the subsequent query.
Therefore, to select documents that have none of a set of words, use:

    !notthis & !northis & !northiseither

Or, using parentheses,

    !(notthis | northis | northiseither)

The keyword B<NOT> (note capitalization) also means negation.  To retrieve
documents containing one word but not another, you can type:

    required AND NOT disallowed

=head2 Phrase Search

To retrieve documents with a specific phrase:

    "this is my phrase"

This will search for documents containing the specified words in sequence,
ignoring punctuation and numbers.

=head2 Proximity

To retrieve documents with words within 10 words of each other:

    first 10 second

or,

    10 first second

The B<NEAR> keyword is an abbreviation for a proximity number of B<10>.
Therefore, the following means the same thing as the two queries above.  Note
that B<NEAR> must be typed all uppercase.

    first NEAR second

Each proximity is evaluated between two adjacent words.  Therefore,

    12 first second third fourth

finds documents containing "second" within 12 words of "first," and "third"
within 12 words of "second," and "fourth" within 12 words of "third."  It does
not find documents with all four words within 12 words of each other.

A proximity number will remain in effect (left to right) until another number
is specified.  A number of "0" will cancel proximity checking.  Note that
proximity searches can be turned off when the index is generated and so may
not be available on all sites.

=head2 Wild Characters

Ferret supports in its queries the same "globbing" characters as is common for
matching filenames (i.e. the familiar "*.*" format).  Such abilities are useful
when searching for such things a the root of a word.  Providing a single "*" as
the last character of the root will match any ending.

    ?     matches any single letter.
    *     matches any zero or more letters.
    [abc] matches a single letter consisting of either an 'a', 'b',
          or 'c.'

To retrieve documents containing a piece of a word:

    onechar? | manychars* | rangechars[aeiou]

Globbing characters are not limited to the end of a word; they may be used
anywhere, even several times within the same word.

=head1 ALPHABETICAL FUNCTION LIST

The following functions are all available from the Ferret module.  Some can be
accessed directly using a prefix of "Ferret::" while others need a Ferret
object ($ferret) that with an open index (see the I<Open> method for more
information.)

===FUNCTIONS===

=head1 COPYRIGHT

B<Ferret> is copyright (c) 1996-2001 by Brian White and Behan Webster.

=head1 SEE ALSO

B<ferret>(1), B<webscan>(1)

=head1 BUGS

Report bugs to: C<bcwhite@pobox.com> S< > Please include the version of
B<Ferret> being run, a detailed description of the problem, and how to
reproduce it.  Thank you!

=cut


###############################################################################
#
# Package-wide variable definitions
#
my %indexmode;
my %newwords;
my %stoppers;
my %options;
my %picsizes;

my $nettimeout	= 0;	# Seconds before aborting net  communications (0=no abt)
my $gettimeout	= 0;	# Seconds before aborting GET  communications (0=no abt)
my $headtimeout	= 0;	# Seconds before aborting HEAD communications (0=no abt)



###############################################################################
#
# Default list of stoppers (two letters or more only)
#
my @defstoppers =
	qw(after all also and any are because before between but can cant could
	   couldnt each for from has have her him how however into its let must
	   mustnt near not other our ours out she should shouldnt since some such
	   than that the them there these they this those under upon use were what
	   when where whether which who why will with within without wont would
	   wouldnt);



###############################################################################
#
# Set of query aliases
#
my %queryaliases =
(
	AND		=> "&",
	OR		=> "|",
	NOT		=> "!",
	NEAR	=> "10",
);



###############################################################################
#
# Required string adjustments for documents & queries
#
sub AliasSubstitute {
	my $usage = 'Usage: Ferret::AliasSubstitute(\$string, \%aliases)';
	@_ == 2 or croak $usage;
	my($query,$aliases) = @_;

	croak $usage if (ref $query   ne "SCALAR");
	croak $usage if (ref $aliases ne "HASH");

#	print "alias in:  $$query\n";

	my($key,$val);
	while (($key,$val) = each %$aliases) {
		$$query =~ s/\b\Q$key\E\b/$val/gs;
	}

#	print "alias out: $$query\n";
}



sub PrepQuery {
	my $usage = 'Usage: Ferret::PrepQuery(\$query [, \$validwords])';
	@_ >= 1 && @_ <= 2 or croak $usage;
	my($query,$wlist) = @_;

	croak $usage if (ref $query ne "SCALAR");
	croak $usage if ($wlist && ref $wlist ne "SCALAR");

	# Remove any dangerous characters
	$$query =~ y/\$/ /s;
#	$$query =~ s/\$/ /g;

	# Remove any unknown characters
	$$query =~ y/A-Za-z0-9\' \t\n!|&()\"\[\]\?\*\^/ /c;
#	$$query =~ s/[^\w\s!|&()\"]+/ /gs;

	# Separate control characters out into words
	$$query =~ s/([!|&()])/ $1 /g;

	# Make sure quoted strings get evaluated properly
	$$query =~ s/"(.*?)"/ ( " $1 " ) /g;

	# Remove global word-matches
	1 while $$query =~ s/(^|\s)(\?|\*|\[\^.*?\])+($|\s)/ /gs;

	# Handle globbing patterns
	foreach (_FindGlobs($$query)) {
		my $expr = "( " . join(' | ',_ExpandGlob($_,$wlist)) . " )";
		$$query =~ s/\Q$_\E/$expr/;
	}

	# Merge all whitespace together
	$$query =~ s/\s+/ /gs;

	# Remove header and footer whitespace
	$$query =~ s/^\s+|\s+$//;

	# Mark entire query as subquery
	$$query = "( $$query )";
}



###############################################################################
#
# Opening & Closing Indexes
#
sub ReadOptions {
	my $usage = 'Usage: $ferret->ReadOptions()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $opts = $this->Options();
	$Ferret::options{$this}	 = $opts;

	if ($opts & &OPT_NOSTOPPERS) {
		$Ferret::stoppers{$this} = "";
	} else {
		$Ferret::stoppers{$this} = $this->DBGetWords("stoppers");
		$Ferret::stoppers{$this} = join("\n",@defstoppers) unless $Ferret::stoppers{$this};
	}
}



sub LocalOpen {
	my $usage = 'Usage: $ferret->LocalOpen(PATHNAME [, WRITEFLAG])';
	@_ >= 2 && @_ <= 3 or croak $usage;
	my($this,$file,$mode) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $open = $this->OpenIndex($file,$mode);

	croak "Error: Could not open index file '$file' -- $!\n       stopped" if $open == -1;
	die	  "Error: Old index incompatible with newer evaluation version -- Please regenerate\n" if $open == -2;
	croak "Error: Evaluation version of \"Ferret\" has expired; stopped" if $open == -3;

	$Ferret::indexmode{$this}= $mode;
	$Ferret::newwords{$this} = "";
	$Ferret::stoppers{$this} = "";

	$this->ReadOptions();
}



=head1 $ferret->Open() -- Open an index for searching

B<Open> must be called before any other "$ferret" methods can function.  It
opens the specified index file, reads any set options, and generally
initializes the index & search systems.

Parameters:
    I<Pathname>         Path to generated index file

Output:
    I<none>

Example:  C<$ferret->>C<Open("ferret.index")>

=cut

sub Open {
	my $usage = 'Usage: $ferret->Open(PATHNAME)';
	@_ == 2 or croak $usage;
	my($this,$file) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->LocalOpen($file);
}



sub FlushLocal {
	my $usage = 'Usage: $ferret->FlushLocal()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	if ($Ferret::indexmode{$this} && $Ferret::newwords{$this}) {
		my $words = $this->DBGetWords("all");
		my @words = split(/\n/, $words);
		push @words,split(/\n/, $Ferret::newwords{$this});
		$words = join(' ',sort(@words));
		$this->DBPutWords("all",$words);
		$Ferret::newwords{$this} = "";
	}
}



=head1 $ferret->Close() -- Close the index

B<Close> should be called when all actions on the index are completed.  Perl
will automatically destruct the object when it goes out of scope, but it's
good style to do these things manually.

Parameters:
    I<none>

Output:
    I<none>

Example:  C<$ferret->>C<Close()>

=cut

sub Close {
	my $usage = 'Usage: $ferret->Close()';
	@_ == 1 or croak $usage;
	my($this) = @_;

	croak $usage if (ref $this ne "Ferret");

	$this->FlushLocal();
	$this->CloseIndex();
}



###############################################################################
#
# Queries
#
sub QueryRun {
	my $usage = 'Usage: $result = $ferret->QueryRun(QUERYSTRING)';
	@_ == 2 or croak $usage;
	my($this,$query) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $wordlist = $this->DBGetWords("all") if $query =~ m/[\[\]\?\*]/;

	# Run standard aliases
	AliasSubstitute(\$query,\%queryaliases);

	StripWords(\$query, \$Ferret::stoppers{$this});
	PrepQuery(\$query,\$wordlist);

#	print "Prepped query=$query\n";

	while ($query =~ m/\(\s*([^()]*?)\s*\)/s) {
		my(@match,$lidx,$ridx,$index);
		my($op,$lr,$quote,$negate,$range);
		$negate = 0;

#		print "Subquery: $1\n";
#		_Time("Parsed Subquery");

		foreach (split(/\s+/,$1)) {
			if ($_ eq '"')		{ $quote = !$quote; next; }
			if ($_ eq '!')		{ $negate= 1;		next; }
			if ($_ =~ m/^\d+$/) { $range = $_;		next; }

			if ($_ =~ m/^[&|]$/) {
				$op = $_ if ($lr);
				next;
			}

			if ($_ =~ m/^\$([-\d]+)$/) {
				$ridx = $1;
			} else {
#				print "Query Word '$_'...";
				$ridx = $this->QueryWord($_);
#				print "(#$ridx)\n";
			}

			if ($negate) {
#				print "Negate Result #$ridx... ";
				$index = $this->QueryNegate($ridx);
#				print "(#$index)\n";
				$this->FreeQueryResult($ridx);
				$ridx = $index;
				$negate = 0;
			}
#			print "Match for '$_': @match\n";

			$range = 1 if $quote;

			if ($lr) {
				if ($op eq '|') {
					$index = $this->QueryOr($lidx,$ridx);
				} elsif ($op eq '' || $op eq '&') {
					$index = $this->QueryAnd($lidx,$ridx,$range);
				} else {
					$@ = "Unknown operator '$op'";
					return;
				}
				$this->FreeQueryResult($lidx);
				$this->FreeQueryResult($ridx);
				$lidx = $index;
			} else {
				$lidx = $ridx;
				$lr++;
			}

			$op = '';
		}

		$query =~ s/\(\s*[^()]*?\s*\)/\$$lidx/;
	}


	my($idx) = ($query =~ m/\$([-\d]+)/);
	return($idx);
}



sub QueryMatches {
	my $usage = 'Usage: @matches = $ferret->QueryMatches($result)';
	@_ == 2 or croak $usage;
	my($this,$idx) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $docs = $this->QueryResults($idx);
	my @docs = split(/\n/,$docs);

	if (@docs) {
		@docs = sort { return $b <=> $a if ($b != $a); $a cmp $b } @docs;
		$@ = "";
	} else {
		$@ = "No matches found";
	}

	return(@docs);
}



=head1 $ferret->Query() -- Search for document by words

B<Query> will search the open index and return an array of matches sorted by
their score.  Each string in the return list has the score prepended to the
name of the matched document and can be separated with a Perl statement such
as:

    my($score,$document) = ($matches[0] =~ m/(\d+) (.*)/);

The query string is quite sophisticated and allows several different operators
between words.

    word    look up given word (may have wild-characters)
    &       boolean AND of the match on left and match on right
    |       boolean OR of the match on left and match on right
    ()      group words and operators to be done first
    !       negate the match on right (i.e. docs "without")
    number  left and right words must be within "number" words

Parameters:
    I<Query-String>     Set of words and operators

Output:
    I<Match-List>       Array of matches, sorted by score

Example:  C<@matches = $ferret->>C<Query("search engine*");>

=cut

sub Query {
	my $usage = 'Usage: @matches = $ferret->Query(QUERYSTRING)';
	@_ == 2 or croak $usage;
	my($this,$query) = @_;

	croak $usage if (ref $this ne "Ferret");

	my $result	= $this->QueryRun($query);
	my @matches = $this->QueryMatches($result);
	$this->FreeQueryResult($result);

	return @matches;
}



1;
__END__
